"""
PV power model
~~~~~~~~~~~~~~

Sources:

{1} Huld, T. et al., 2010. Mapping the performance of PV modules,
    effects of module type and data averaging. Solar Energy, 84(2),
    p.324-338. DOI: 10.1016/j.solener.2009.12.002


General assumptions made:
-------------------------

Maximum power point tracking is assumed.

"""

import math
import warnings

import numpy as np
import pandas as pd

from gsee import trigon

# Constants
R_TAMB = 20  # Reference ambient temperature (degC)
R_TMOD = 25  # Reference module temperature (degC)
R_IRRADIANCE = 1  # Reference irradiance (kW/m2)


class PVPanel(object):
    """
    PV panel model class

    Unit for power is kW, for energy, kWh.

    By default, self.module_aperture is set to 1.0, so the output will
    correspond to output per m2 of solar field given the other
    input values.

    Parameters
    ----------
    panel_aperture : float
        Panel aperture area (m2)
    ref_efficiency : float
        Reference conversion efficiency
    c_temp_amb: float, default 1 degC / degC
        Panel temperature coefficient of ambient temperature
    c_temp_irrad: float, default 35 degC / (kW/m2)
        Panel temperature coefficient of irradiance. According to {1},
        reasonable values for this for c-Si are:
            35  # Free-standing module, assuming no wind
            50  # Building-integrated module

    """
    def __init__(self, panel_aperture=1.0, panel_ref_efficiency=1.0,
                 c_temp_amb=1, c_temp_irrad=35):
        super().__init__()
        # Panel characteristics
        self.panel_aperture = panel_aperture
        self.panel_ref_efficiency = panel_ref_efficiency
        self.use_diffuse = True  # Whether to also use diffuse irradiance
        # Panel temperature estimation
        self.c_temp_tamb = c_temp_amb
        self.c_temp_irrad = c_temp_irrad

    def panel_power(self, direct, diffuse=None, tamb=None):
        """
        Returns power in kW from PV panel(s) based on given input data.

        Parameters
        ----------
        direct : pandas Series
            Direct irradiance hitting the panel(s) in kW/m2.
        diffuse : pandas Series, default None
            Diffuse irradiance hitting the panel(s) in kW/m2.
        tamb : pandas Series, default None
            Ambient temperature in deg C. If not given, R_TAMB is used
            for all values.

        """
        index_msg = 'Data indices must match'
        if diffuse is not None:
            assert direct.index.equals(diffuse.index), index_msg
        if tamb is not None:
            assert direct.index.equals(tamb.index), index_msg
        if self.use_diffuse:
            irradiance = direct + diffuse
        else:
            irradiance = direct
        return (irradiance * self.panel_aperture
                * self.panel_relative_efficiency(irradiance, tamb)
                * self.panel_ref_efficiency)

    def panel_relative_efficiency(self, irradiance, tamb):
        """
        Returns the relative conversion efficiency modifier as a
        function of irradiance and ambient temperature.

        Source: {1}

        Parameters
        ----------
        irradiance : pandas Series
            Irradiance in kW
        tamb : pandas Series
            Ambient temperature in deg C

        """
        # G_: normalized in-plane irradiance
        G_ = irradiance / R_IRRADIANCE
        # T_: normalized module temperature
        T_ = (self.c_temp_tamb * tamb + self.c_temp_irrad * irradiance) - R_TMOD
        # NB: np.log without base implies base e or ln
        # Catching warnings to suppress "RuntimeWarning: invalid value encountered in log"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            eff = (
                1 + self.k_1 * np.log(G_)
                + self.k_2 * (np.log(G_)) ** 2
                + T_ * (self.k_3
                        + self.k_4 * np.log(G_)
                        + self.k_5 * (np.log(G_)) ** 2)
                + self.k_6 * (T_ ** 2)
            )
        eff.fillna(0, inplace=True)  # NaNs in case that G_ was <= 0
        eff[eff < 0] = 0  # Also make sure efficiency can't be negative
        return eff


class CSiPanel(PVPanel):
    """c-Si technology, based on data from {1}"""
    def __init__(self, **kwargs):
        super(CSiPanel, self).__init__(**kwargs)
        self.k_1 = -0.017162
        self.k_2 = -0.040289
        self.k_3 = -0.004681
        self.k_4 = 0.000148
        self.k_5 = 0.000169
        self.k_6 = 0.000005


class CdTePanel(PVPanel):
    """CdTe technology, based on data from {1}"""
    def __init__(self, **kwargs):
        super(CdTePanel, self).__init__(**kwargs)
        self.k_1 = -0.103251
        self.k_2 = -0.040446
        self.k_3 = -0.001667
        self.k_4 = -0.002075
        self.k_5 = -0.001445
        self.k_6 = -0.000023


_PANEL_TYPES = {
    'csi': CSiPanel,
    'cdte': CdTePanel
}


def run_model(
        data, coords, tilt, azim, tracking, capacity,
        technology='csi', system_loss=0.10, angles=None,
        include_raw_data=False, **kwargs):
    """
    Run PV plant model.

    Parameters
    ----------
    data : pandas DataFrame
        Must contain columns 'global_horizontal' (in kW/m2)
        and 'diffuse_fraction', and may contain a 'temperature' column
        for ambient air temperature (in deg C).
    coords : (float, float) tuple
        Latitude and longitude.
    tilt : float
        Tilt angle (degrees).
    azim : float
        Azimuth angle (degrees, 180 = towards equator).
    tracking : int
        Tracking (0: none, 1: 1-axis, 2: 2-axis).
    capacity : float
        Installed capacity in kW.
    technology : str, default 'csi'
        Panel technology, must be one of 'csi', 'cdte', 'cpv'
    system_loss : float, default 0.10
        Total system power losses (fraction).
    angles : pandas DataFrame, default None
        Solar angles. Ff already computed, speeds up the computations.
    include_raw_data : bool, default False
        If true, returns a DataFrame instead of Series which includes
        the input data (panel irradiance and temperature).
    kwargs : additional kwargs passed on the model constructor

    Returns
    -------
    result : pandas Series
        The PV system output in kW for each hour.

    """
    if (system_loss < 0) or (system_loss > 1):
        raise ValueError('system_loss must be >=0 and <=1')

    # Process data
    dir_horiz = data.global_horizontal * (1 - data.diffuse_fraction)
    diff_horiz = data.global_horizontal * data.diffuse_fraction
    # TODO more flexibilty when passing in data, e.g. allow passing in
    # other combinations of data like DNI + global horizontal
    # NB: aperture_irradiance expects azim/tilt in radians!
    irrad = trigon.aperture_irradiance(dir_horiz, diff_horiz, coords,
                                       tracking=tracking,
                                       azimuth=math.radians(azim),
                                       tilt=math.radians(tilt),
                                       angles=angles)
    datetimes = irrad.index

    # Temperature, if it was given
    if 'temperature' in data.columns:
        tamb = data['temperature']
    else:
        tamb = pd.Series(R_TAMB, index=datetimes)

    # Set up the panel model
    # NB: panel efficiency is not used here, but we retain the possibility
    # to adjust both efficiency and panel size in case we want to emulate
    # specific panel types
    panel_class = _PANEL_TYPES[technology]
    panel_efficiency = 0.1
    area_per_capacity = 1 / panel_efficiency

    # TODO allow panel_aperture to change through day as shading takes place
    panel = panel_class(panel_aperture=capacity * area_per_capacity,
                        panel_ref_efficiency=panel_efficiency,
                        **kwargs)

    # Run the panel model and return output
    # TODO this ignores details like shading due to close panels
    output = panel.panel_power(irrad.direct,
                               irrad.diffuse,
                               tamb)
    sim = pd.Series(output, index=datetimes).clip_upper(capacity) * (1 - system_loss)
    if include_raw_data:
        items = [
            ('output', sim),
            ('direct', irrad.direct),
            ('diffuse', irrad.diffuse),
            ('temperature', tamb)
        ]
        return pd.DataFrame.from_items(items)
    else:
        return sim


def optimal_tilt(lat):
    """
    Returns an optimal tilt angle for the given ``lat``, assuming that
    the panel is facing towards the equator, using a simple method from [1].

    This method only works for latitudes between 0 and 50. For higher
    latitudes, a static 40 degree angle is returned.

    These results should be used with caution, but there is some
    evidence that tilt angle may not be that important [2].

    [1] http://www.solarpaneltilt.com/#fixed
    [2] http://dx.doi.org/10.1016/j.solener.2010.12.014

    Parameters
    ----------
    lat : float
        Latitude in degrees.

    Returns
    -------
    angle : float
        Optimal tilt angle in degrees.

    """
    lat = abs(lat)
    if lat <= 25:
        return lat * 0.87
    elif lat <= 50:
        return (lat * 0.76) + 3.1
    else:  # lat > 50
        # raise NotImplementedError('Not implemented for latitudes beyond 50.')
        return 40  # Simply use 40 degrees above lat 50



"""

An implementation of the `BRL` diffuse solar fraction model as described in
Ridley et al. (2010):

http://dx.doi.org/10.1016/j.renene.2009.07.018

Includes updated model parameters estimated via Bayesian inference in
Lauret et al. (2013):

http://dx.doi.org/10.1016/j.renene.2012.01.049

"""

import datetime
import math

import ephem
import pandas as pd
import numpy as np

from gsee import trigon


def _solartime(observer, sun):
    """Return solar time for given observer and sun"""
    # sidereal time == ra (right ascension) is the highest point (noon)
    hour_angle = observer.sidereal_time() - sun.ra
    return ephem.hours(hour_angle + ephem.hours('12:00')).norm  # norm for 24h


def _get_psi_func(sunrise, sunset):
    """
    Return a function, psi(hour, ks), for the given sunrise and
    sunset times

    """
    try:
        sunrise_hour = sunrise.hour
    except AttributeError:
        sunrise_hour = 0
    try:
        sunset_hour = sunset.hour
    except AttributeError:
        sunset_hour = 23

    def f(hour, ks):
        if (hour > sunrise_hour) and (hour < sunset_hour):
            psi = (ks[hour - 1] + ks[hour + 1]) / 2
            # Extra check: in some cases there is no data in `ks` even before
            # sunset / afer sunrise. For example, if the sun sets just minutes
            # after the hour, there may be no irradiance data in that hour.
            # This if-clause ensures that for practical reasons such cases
            # are treated as if they were sunrise/sunset hours.
            if np.isnan(psi):
                if np.isnan(ks[hour - 1]):
                    psi = ks[hour + 1]
                else:
                    psi = ks[hour - 1]
        elif hour == sunrise_hour:
            try:
                psi = ks[hour + 1]
            except IndexError:
                psi = ks[hour]
        elif hour == sunset_hour:
            try:
                psi = ks[hour - 1]
            except IndexError:
                psi = ks[hour]
        else:
            psi = 0
        return psi
    return f


# Updated params from Lauret et al. (2013)
DEFAULT_PARAMS = {'a0': -5.32,
                  'a1': 7.28,
                  'b1': -0.03,
                  'b2': -0.0047,
                  'b3': 1.72,
                  'b4': 1.08}


# Parameters from Ridley et al. (2010)
# DEFAULT_PARAMS = {'a0': -5.38,
#                   'a1': 6.63,
#                   'b1': 0.006,
#                   'b2': -0.007,
#                   'b3': 1.75,
#                   'b4': 1.31}


# TODO make this into a get_daily_diffuse_func with p param,
# so that can easily switch the predictor parameters!
def _daily_diffuse(obs, ks, sunrise, sunset, p=DEFAULT_PARAMS):
    """
    Returns a list of diffuse fractions for the given observer
    which must have its coordinates and date set, and given the ``ks``,
    a list of 24 hourly clearness indices, and sunrise and sunset times.

    """
    date = obs.date.datetime()
    # whether date was set or not, ensure it's at hour 0
    obs.date = datetime.datetime(date.year, date.month, date.day)
    sun = ephem.Sun()
    sun.compute(obs)
    # sunrise, sunset = _sunrise_sunset(obs, sun)
    alpha = sun.alt
    values = []
    k_day = pd.Series(ks).mean()  # using pandas to ignore NaN
    psi = _get_psi_func(sunrise, sunset)
    for hour in range(24):
        if np.isnan(ks[hour]):
            d = np.nan
        else:
            ast = _solartime(obs, sun)
            pwr = (p['a0'] + p['a1'] * ks[hour]
                   + p['b1'] * ast + p['b2'] * alpha
                   + p['b3'] * k_day + p['b4'] * psi(hour, ks))
            d = 1 / (1 + math.e ** pwr)
        values.append(d)
        # Increase obs.date by one hour for the next iteration
        obs.date = obs.date.datetime() + datetime.timedelta(hours=1)
        sun.compute(obs)
    return values


def run(hourly_clearness, coords, rise_set_times=None):
    """Run the BRL model

    Parameters
    ----------
    hourly_clearness : pandas Series
        Hourly clearness indices with a datetime index.
    coords : 2-tuple of floats or ints
        Latitude and longitude
    rise_set_times : list
        List of (sunrise, sunset) time tuples, if not given, is
        calculated here.

    Returns
    -------
    result : pandas Series
        Diffuse fractions with the same datetime index as hourly_clearness.

    """
    obs = ephem.Observer()
    obs.lat = str(coords[0])
    obs.lon = str(coords[1])
    if not rise_set_times:
        rise_set_times = trigon._sun_rise_set(hourly_clearness.index, obs)
    diffuse_fractions = []
    for i in range(0, len(hourly_clearness), 24):
        # for entry in list in hourly clearness indices:
        ks = hourly_clearness.iloc[i:i+24].tolist()
        obs.date = hourly_clearness.index[i]
        # These are indexed by day, so need to scale the index
        sunrise, sunset = rise_set_times[int(i / 24)]
        results = _daily_diffuse(obs, ks, sunrise, sunset)
        diffuse_fractions.extend(results)
    return pd.Series(diffuse_fractions, index=hourly_clearness.index)



"""
Irradiance on an inclined plane
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Using trigonometry (Lambert's cosine law, etc).

"""

import datetime

import ephem
import numpy as np
import pandas as pd


def _sun_rise_set(datetime_index, obs):
    """
    Returns a list of (sunrise, sunset) tuples for the datetimes
    in the given datetime_index, assuming that the index is daily.

    Sunrise or sunset will be None if the sun doesn't rise/set.

    """
    sun = ephem.Sun()
    times = []

    def _get_datetime(date):
        obs.date = date
        sun.compute(obs)
        if sun.rise_time is None:
            rise_time = None
        else:
            rise_time = sun.rise_time.datetime()
        if sun.set_time is None:
            set_time = None
        else:
            set_time = sun.set_time.datetime()
        return (rise_time, set_time)

    for i in datetime_index:
        times.append(_get_datetime(i))
    return times


def sun_rise_set_times(datetime_index, coords):
    """
    Return sunrise and set times for the given datetime_index and coords.

    The datetime_index will be resampled to daily frequency.

    """
    obs = ephem.Observer()
    obs.lat = str(coords[0])
    obs.lon = str(coords[1])
    # Ensure datetime_index is daily
    dtindex = pd.DatetimeIndex(datetime_index.to_series().map(pd.Timestamp.date).unique())
    return _sun_rise_set(dtindex, obs)


def sun_angles(datetime_index, coords, rise_set_times=None):
    """Calculate sun angles. Returns a dataframe containing `sun_alt`,
    `sun_zenith`, `sun_azimuth` and `duration` over the passed datetime index.

    Parameters
    ----------
    datetime_index : pandas datetime index
        Handled as if they were UTC not matter what timezone info
        they may supply.
    coords : (float, float) or (int, int) tuple
        Latitude and longitude.
    rise_set_times : list, default None
        List of (sunrise, sunset) time tuples, if not passed, is computed
        here.

    """

    def _sun_alt_azim(sun, obs):
        sun.compute(obs)
        return sun.alt, sun.az

    # Initialize ephem objects
    obs = ephem.Observer()
    obs.lat = str(coords[0])
    obs.lon = str(coords[1])
    sun = ephem.Sun()

    # Calculate daily sunrise/sunset times
    if not rise_set_times:
        rise_set_times = _sun_rise_set(datetime_index, obs)

    # Calculate hourly altitute, azimuth, and sunshine
    alts = []
    azims = []
    durations = []

    for index, item in enumerate(datetime_index):
        obs.date = item
        # rise/set times are indexed by day, so need to scale the index
        rise_time, set_time = rise_set_times[int(index / 24)]

        # Set angles, sun altitude and duration based on hour of day:
        if rise_time is not None and item.hour == rise_time.hour:
            # Special case for sunrise hour
            duration = 60 - rise_time.minute - (rise_time.second / 60.0)
            obs.date = rise_time + datetime.timedelta(minutes=duration / 2)
            sun_alt, sun_azimuth = _sun_alt_azim(sun, obs)
        elif set_time is not None and item.hour == set_time.hour:
            # Special case for sunset hour
            duration = set_time.minute + set_time.second / 60.0
            obs.date = item + datetime.timedelta(minutes=duration / 2)
            sun_alt, sun_azimuth = _sun_alt_azim(sun, obs)
        else:
            # All other hours
            duration = 60
            obs.date = item + datetime.timedelta(minutes=30)
            sun_alt, sun_azimuth = _sun_alt_azim(sun, obs)
            if sun_alt < 0:  # If sun is below horizon
                sun_alt, sun_azimuth, duration = 0, 0, 0

        alts.append(sun_alt)
        azims.append(sun_azimuth)
        durations.append(duration)
    df = pd.DataFrame({'sun_alt': alts, 'sun_azimuth': azims,
                       'duration': durations},
                      index=datetime_index)
    df['sun_zenith'] = (np.pi / 2) - df.sun_alt
    return df


def _incidence_fixed(sun_alt, tilt, azimuth, sun_azimuth):
    return np.arccos(np.sin(sun_alt) * np.cos(tilt)
                     + np.cos(sun_alt) * np.sin(tilt)
                     * np.cos(azimuth - sun_azimuth))


def _incidence_single_tracking(sun_alt, tilt, azimuth, sun_azimuth):
    if tilt == 0:
        return np.arccos(np.sqrt(1 - np.cos(sun_alt) ** 2
                         * np.cos(sun_azimuth - azimuth) ** 2))
    else:
        return np.arccos(np.sqrt(1 - (np.cos(sun_alt + tilt) * np.cos(tilt)
                         * np.cos(sun_alt) * (1 - np.cos(sun_azimuth
                                                         - azimuth))) ** 2))


def _tilt_single_tracking(sun_alt, tilt, azimuth, sun_azimuth):
    if tilt == 0:
        return np.arctan(np.sin(sun_azimuth - azimuth) / np.tan(sun_alt))
    else:
        return np.arctan((np.cos(sun_alt) * np.sin(sun_azimuth - azimuth))
                         / (np.sin(sun_alt - tilt) + np.sin(tilt)
                            * np.cos(sun_alt) * (1 - np.cos(sun_azimuth)
                                                 - azimuth)))


def aperture_irradiance(direct, diffuse, coords,
                        tilt=0, azimuth=0, tracking=0, albedo=0.3,
                        dni_only=False, angles=None):
    """
    Args:
        direct : a series of direct horizontal irradiance with a datetime index
        diffuse : a series of diffuse horizontal irradiance with the same
                  datetime index as for direct
        coords : (lat, lon) tuple of location coordinates
        tilt : angle of panel relative to the horizontal plane, 0 = flat
        azimuth : deviation of the tilt direction from the meridian,
                  0 = towards pole, going clockwise, 3.14 = towards equator
        tracking : 0 (none, default), 1 (tilt), or 2 (tilt and azimuth).
                   If 1, azimuth is the orientation of the tilt axis, which
                   can be horizontal (tilt=0) or tilted.
        albedo : reflectance of the surrounding surface
        dni_only : only calculate and directly return a DNI time series
                   (ignores tilt, azimuth, tracking and albedo arguments)
        angles : solar angles, if default (None), is computed here

    """
    # 0. Correct azimuth if we're on southern hemisphere, so that 3.14
    # points north instead of south
    if coords[0] < 0:
        azimuth = azimuth + np.pi
    # 1. Calculate solar angles
    if angles is None:
        sunrise_set_times = sun_rise_set_times(direct.index, coords)
        angles = sun_angles(direct.index, coords, sunrise_set_times)
    # 2. Calculate direct normal irradiance
    dni = (direct * (angles['duration'] / 60)) / np.cos(angles['sun_zenith'])
    if dni_only:
        return dni
    # 3. Calculate appropriate aperture incidence angle
    if tracking == 0:
        incidence = _incidence_fixed(angles['sun_alt'], tilt, azimuth,
                                     angles['sun_azimuth'])
        panel_tilt = tilt
    elif tracking == 1:
        # 1-axis tracking with horizontal or tilted tracking axis
        incidence = _incidence_single_tracking(angles['sun_alt'],
                                               tilt, azimuth,
                                               angles['sun_azimuth'])
        panel_tilt = _tilt_single_tracking(angles['sun_alt'], tilt, azimuth,
                                           angles['sun_azimuth'])
    elif tracking == 2:
        # 2-axis tracking means incidence angle is zero
        # Assuming azimuth/elevation tracking for tilt/azimuth angles
        incidence = 0
        panel_tilt = angles['sun_zenith']
        azimuth = angles['sun_azimuth']
    else:
        raise ValueError('Invalid setting for tracking: {}'.format(tracking))
    # 4. Compute direct and diffuse irradiance on plane
    # clip_lower(0) ensures that very low panel to sun altitude angles do not
    # result in negative direct irradiance (reflection)
    plane_direct = (dni * np.cos(incidence)).fillna(0).clip_lower(0)
    plane_diffuse = (diffuse * ((1 + np.cos(panel_tilt)) / 2)
                     + albedo * (direct + diffuse)
                     * ((1 - np.cos(panel_tilt)) / 2)).fillna(0)
    return pd.DataFrame({'direct': plane_direct, 'diffuse': plane_diffuse})



#!/usr/bin/env python

from setuptools import setup, find_packages

# Sets the __version__ variable
exec(open('gsee/_version.py').read())

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='gsee',
    version=__version__,
    author='Stefan Pfenninger',
    author_email='stefan@pfenninger.org',
    description='GSEE: Global Solar Energy Estimator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/renewables-ninja/gsee',
    packages=find_packages(),
    install_requires=[
        "pyephem >= 3.7.6.0",
        "numpy >= 1.10.1",
        "pandas >= 0.18.0"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)
